#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "ply.h"

typedef struct PlyVertex {
  float x,y,z;
} PlyVertex;

typedef struct PlyFace {
  unsigned char nverts;
  int *verts;
} PlyFace;

PlyProperty vertex_props[] = { /* list of property information for a vertex */
  {"x", Float32, Float32, offsetof(PlyVertex,x), 0, 0, 0, 0},
  {"y", Float32, Float32, offsetof(PlyVertex,y), 0, 0, 0, 0},
  {"z", Float32, Float32, offsetof(PlyVertex,z), 0, 0, 0, 0},
};

PlyProperty face_props[] = { /* list of property information for a face */
  {"vertex_indices", Int32, Int32, offsetof(PlyFace,verts), 1, Uint8, Uint8, offsetof(PlyFace,nverts)},
};

/* vertex and face definitions */

class Permute
{
public:
	int p;
};

class Vertex : public Permute
{
public:
	float coords[3];
};

class Face : public Permute
{
public:
  int index[3];
};

///// functions used during bucket sort

void insert_into_bucket1(int* bucket1, int* twininv, int bucket, int pointer)
{
	twininv[pointer] = bucket1[bucket];
	bucket1[bucket] = pointer;
}

void insert_into_bucket2(int* bucket2, int* twininv, int bucket, int pointer)
{
	twininv[pointer] = bucket2[bucket];
	bucket2[bucket] = pointer;
}

///// functions to traverse the twinedge structure

#define ENCODED 0x80000000
#define ENCODED_VERTEX 0x40000000
#define MASK 0x3FFFFFFF

// get the next edge
int next(int e)
{
	return (3 * (e / 3) + ((e + 1) % 3));
}

// get the prev edge
int prev(int e)
{
	return (3 * (e / 3) + ((e + 2) % 3));
}

// get the origin of the edge

#define inv(e) (twininv[e])
#define origin(e) (twinorigin[e] & MASK)
#define markTwinedgeEncoded(e) (twinorigin[e] |= ENCODED)
#define markVertexEncoded(e) (twinorigin[e] |= ENCODED_VERTEX)
#define encodedTwinedge(e) (twinorigin[e] & ENCODED)
#define encodedVertex(e) (twinorigin[e] & ENCODED_VERTEX)

int gate_stack_entries = 0;
int gate_stack[10000];

int pop_gate()
{
  if (gate_stack_entries)
  {
    return gate_stack[--gate_stack_entries];
  }
  else
  {
    return -1;
  }
}

void push_gate(int gate)
{
  gate_stack[gate_stack_entries++] = gate;
}

bool load_mesh(char* file_name, Vertex** verts_out, int* nverts_out, Face** faces_out,  int* nfaces_out)
{
  int i,j;
  int elem_count;
  char* elem_name;
  PlyFile *in_ply = open_ply_for_read(file_name);

  for (i = 0; i < in_ply->num_elem_types; i++)
	{
	  elem_name = setup_element_read_ply (in_ply, i, &elem_count);

		if (equal_strings ("vertex", elem_name))
		{
      /* create the memory for the verts */
      Vertex* verts = new Vertex[elem_count]; 

			/* set up for getting vertex elements */
			setup_property_ply (in_ply, &vertex_props[0]);
			setup_property_ply (in_ply, &vertex_props[1]);
			setup_property_ply (in_ply, &vertex_props[2]);
			
			PlyVertex* vertex = (PlyVertex *) malloc (sizeof (PlyVertex));
			/* grab all the vertex elements */
			for (j = 0; j < elem_count; j++)
			{
				get_element_ply (in_ply, (void *)vertex);
				verts[j].coords[0] = vertex->x;
				verts[j].coords[1] = vertex->y;
				verts[j].coords[2] = vertex->z;
			}
			free(vertex);
      *verts_out = verts;
      *nverts_out = elem_count;
		}
		else if (equal_strings ("face", elem_name))
		{
      /* create the memory for the verts */
      Face* faces = new Face[elem_count]; 

			/* set up for getting face elements */
      setup_property_ply (in_ply, &face_props[0]);

			/* grab all the face elements */
			PlyFace* face = (PlyFace*) malloc(sizeof(PlyFace));
      for (j = 0; j < elem_count; j++)
			{
        get_element_ply (in_ply, (void *) face);
        faces[j].index[0] = face->verts[0];
        faces[j].index[1] = face->verts[1];
        faces[j].index[2] = face->verts[2];
				free(face->verts);
      }
			free(face);
      *faces_out = faces;
      *nfaces_out = elem_count;
    }
  }
  close_ply (in_ply);
  free_ply (in_ply);

  return true;
}

bool construct_connectivity(int nfaces, Face* faces, int nverts, int** twinorigin_out,  int** twininv_out)
{
  int i,j;
  Face* face;
  int ntwinedges = 3*nfaces;

  // create first bucket
  int* bucket1 = (int*)malloc(sizeof(int)*nverts);

	if (bucket1 == 0)
	{
		printf("malloc for bucket1 failed\n");
		exit(0);
  }

	for (i = 0; i < nverts; i++)
	{
		bucket1[i] = -1;
	}
  
  // create edges

	int* twinorigin = new int[ntwinedges];
	if (twinorigin == 0)
	{
		printf("new for twinorigin failed\n");
		exit(0);
	}

  int* twininv = new int[ntwinedges];
	if (twininv == 0)
	{
		printf("new for twininv failed\n");
		exit(0);
	}
  
  /* set up for getting face elements */

  for (j = 0; j < nfaces; j++)
	{
    face = &(faces[j]);
	  // edge from face->index[0] -> face->index[1]
	  if (face->index[0] < face->index[1])
	  {
		  insert_into_bucket1(bucket1, twininv, face->index[0], 3*j+0);
		  twinorigin[3*j+0] = face->index[1];
	  }
	  else
	  {
		  insert_into_bucket1(bucket1, twininv, face->index[1], 3*j+0);
		  twinorigin[3*j+0] = -(face->index[0]+1);
	  }

	  // edge from face->index[1] -> face->index[2]
	  if (face->index[1] < face->index[2])
	  {
		  insert_into_bucket1(bucket1, twininv, face->index[1], 3*j+1);
		  twinorigin[3*j+1] = face->index[2];
	  }
	  else
	  {
		  insert_into_bucket1(bucket1, twininv, face->index[2], 3*j+1);
		  twinorigin[3*j+1] = -(face->index[1]+1);
	  }

	  // edge from face->index[2] -> face->index[0]
	  if (face->index[2] < face->index[0])
	  {
		  insert_into_bucket1(bucket1, twininv, face->index[2], 3*j+2);
		  twinorigin[3*j+2] = face->index[0];
	  }
	  else
	  {
		  insert_into_bucket1(bucket1, twininv, face->index[0], 3*j+2);
		  twinorigin[3*j+2] = -(face->index[2]+1);
	  }
  }

	// create second bucket
			
	int* bucket2 = (int*)malloc(sizeof(int)*nverts);

	if (bucket2 == 0)
	{
		printf("malloc for bucket2 failed\n");
		exit(1);
	}

	for (i = 0; i < nverts; i++)
	{
		bucket2[i] = -1;
	}

  int pointer, next_pointer;

	for (i = 0; i < nverts; i++)
	{
		pointer = bucket1[i];

		while (pointer != -1)
		{
			next_pointer = twininv[pointer];
			if (twinorigin[pointer] >= 0)
			{
				insert_into_bucket2(bucket2, twininv, twinorigin[pointer], pointer);
				twinorigin[pointer] = i;
			}
			else
			{
				insert_into_bucket2(bucket2, twininv, -1-twinorigin[pointer], pointer);
				twinorigin[pointer] = -(i+1);
			}
			pointer = next_pointer;
		}
  }

	free(bucket1);
	
	/***********************/
	/* create connectivity */
	/***********************/

  int e0, e0val;
  int e1, e1val;
  int e2, e2val;
  int num_edges = 0;

	for (i = 0; i < nverts; i++)
	{
		pointer = bucket2[i];
		while (pointer != -1)
		{
			e0 = pointer;
			if (e0 != -1)
			{
				e0val = (twinorigin[e0] >= 0) ? twinorigin[e0] : -1-twinorigin[e0];
			}
			e1 = twininv[e0];
			if (e1 != -1)
			{
				e1val = (twinorigin[e1] >= 0) ? twinorigin[e1] : -1-twinorigin[e1];
			}
			e2 = (e1 != -1) ? twininv[e1] : -1;
			if (e2 != -1)
			{
				e2val = (twinorigin[e2] >= 0) ? twinorigin[e2] : -1-twinorigin[e2];
			}

		  if ((e1 != -1) && (e0val == e1val))
			{
				if ( (twinorigin[e0] + twinorigin[e1] == -1) && ((e2 == -1) || (e0val != e2val)) ) // manifold & oriented?
				{
          num_edges++;
					twininv[e1] = e0;
					twininv[e0] = e1;
					if (twinorigin[e0] >= 0)
					{
						twinorigin[e1] = i; // set origin
					}
					else
					{
						twinorigin[e0] = i; // set origin
					}
					pointer = e2;
				}
				else
				{
        	printf("non_manifold or not_oriented case not supported\n");
          exit(1);
				}
			}
			else
			{
        printf("border edge case not supported\n");
        exit(1);
      }
		}
	}
	free(bucket2);

//  printf("matched %d edges\n",num_edges);

  *twinorigin_out = twinorigin;
  *twininv_out = twininv;

	return true;
}

bool compute_reference_ordering(int nverts, Vertex* verts, Vertex** verts_reference, int nfaces, Face* faces, Face** faces_reference, int* twinorigin, int* twininv)
{
  int v_count = 0;
  int f_count = 0;

  // put first face and first thee vertices into reference arrays
  verts_reference[v_count++] = &(verts[origin(0)]);
  verts_reference[v_count++] = &(verts[origin(1)]);
  verts_reference[v_count++] = &(verts[origin(2)]);
  faces_reference[f_count++] = &(faces[0/3]);

  // mark as encoded
  markVertexEncoded(origin(0));
  markVertexEncoded(origin(1));
  markVertexEncoded(origin(2));
  markTwinedgeEncoded(inv(0));
  markTwinedgeEncoded(inv(1));
  markTwinedgeEncoded(inv(2));

  int redge;
  int ledge;
  int gate = inv(1);

  while (true)
  {
    faces_reference[f_count++] = &(faces[gate/3]);
    redge = next(gate);
    ledge = next(redge);
    if (encodedTwinedge(redge))
    {
      if (encodedTwinedge(ledge))
      {
        gate = pop_gate();
        if (gate == -1) break;
      }
      else
      {
        markTwinedgeEncoded(inv(ledge));
        gate = inv(ledge);
      }
    }
    else if (encodedTwinedge(ledge))
    {
      markTwinedgeEncoded(inv(redge));
      gate = inv(redge);
    }
    else
    {
      markTwinedgeEncoded(inv(redge));
      markTwinedgeEncoded(inv(ledge));
      if (encodedVertex(origin(ledge)))
      {
        push_gate(inv(ledge));
      }
      else
      {
        verts_reference[v_count++] = &(verts[origin(ledge)]);
        markVertexEncoded(origin(ledge));
      }
      gate = inv(redge);
    }
  }

//  printf("reordered %d of %d and %d of %d\n",v_count, nverts, f_count, nfaces);

  return true;
}

/*
int main(int argc, char** argv)
{
  Vertex* verts;
  int nverts;
  Face* faces;
  int nfaces;

  load_mesh("horse.ply", &verts, &nverts, &faces, &nfaces);

  int* twinorigin;
  int* twininv;

  construct_connectivity(nfaces, faces, nverts, &twinorigin,  &twininv);

  Vertex** verts_reference = new Vertex*[nverts];
  Face** faces_reference = new Face*[nfaces];

  compute_reference_ordering(nverts, verts, verts_reference, nfaces, faces, faces_reference, twinorigin, twininv);

  return 0;
}
*/