/*
===============================================================================

  FILE:  stega_mesh.cpp
  
  CONTENTS:
  
    A simple algorithm that encodes/decodes a random bitstring as a permutation
		of elements from a given reference ordering in O(n) time and O(n) space.

    A somewhat more complete example that "stega.cpp" but also this is just a
    proof-of-concept implementation that only works with perfectly manifold
    meshes without boundary and with sphere topology (e.g. genus zero). 
  
  PROGRAMMERS:
  
    martin isenburg@cs.unc.edu
  
  COPYRIGHT:
  
    copyright (C) 2007 martin isenburg@cs.unc.edu
    
    This software is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  
  CHANGE HISTORY:
  
    24 December 2007 -- adapted from stega.cpp for "real" timings with meshes
  
===============================================================================
*/

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <time.h>

class Permute
{
public:
	int p;
};

class Vertex : public Permute
{
public:
	float coords[3];
};

class Face : public Permute
{
public:
  int index[3];
};

bool load_mesh(char* file_name, Vertex** verts_out, int* nverts_out, Face** faces_out,  int* nfaces_out);
bool construct_connectivity(int nfaces, Face* faces, int nverts, int** twinorigin_out,  int** twininv_out);
bool compute_reference_ordering(int nverts, Vertex* verts, Vertex** verts_reference, int nfaces, Face* faces, Face** faces_reference, int* twinorigin, int* twininv);

static int bits_in_bit_buffer1;
static unsigned int bit_buffer1;
static unsigned int bit_buffer2;
static unsigned int bit_count;

static void init_bits()
{
	srand(0);
	bits_in_bit_buffer1 = 32;
	bit_buffer1 = ((rand()&2047) << 21) | ((rand()&2047) << 10) | (rand()&1023);
	bit_buffer2 = ((rand()&2047) << 21) | ((rand()&2047) << 10) | (rand()&1023);
	bit_count = 0;
}

static unsigned int peek_bits(int bits)
{
	int result;
	assert(bits <= 32 && bits_in_bit_buffer1);
	if (bits_in_bit_buffer1 >= bits)
	{
		result = bit_buffer1 & ((1<<bits)-1);
	}
	else
	{
		result = bit_buffer1 | ((bit_buffer2 & ((1<<(bits-bits_in_bit_buffer1))-1)) << bits_in_bit_buffer1);
	}
	return result;
}

static void pop_bits(int bits)
{
	assert(bits <= 32 && bits_in_bit_buffer1);
	if (bits_in_bit_buffer1 > bits)
	{
		bits_in_bit_buffer1 -= bits;
		bit_buffer1 >>= bits;
	}
	else
	{
		bits_in_bit_buffer1 = bits_in_bit_buffer1-bits+32;
		bit_buffer1 = bit_buffer2 >> (32-bits_in_bit_buffer1);
		bit_buffer2 = ((rand()&2047) << 21) | ((rand()&2047) << 10) | (rand()&1023);
	}
	bit_count += bits;
}

static int encode(Permute** reference, Permute** permutation, int nelements)
{
	int bit_count_start = bit_count;

	// compute k and when k will decrement
	int i = nelements;
	int k = 0;
	while (i/=2) k++;
	int twopowk = 1<<k;
	
	// encode either k+1 or k bits with each element
	int p;
  for (i = 0; i < nelements-1; i++)
	{
		p = peek_bits(k+1);
		if (p >= twopowk && p < nelements-i)
		{
			// we were able to encode k+1 bits
			pop_bits(k+1);
		}
		else
		{
			// can only encode k bits
			p = peek_bits(k);
			pop_bits(k);
		}
		// we encode them by putting the reference element p in the output
		permutation[i] = reference[p];
    // this line replaces the binary tree
		// the last element of the reference array becomes the new reference element p 
		reference[p] = reference[nelements-1-i];
		// is k becomming smaller
		if (nelements-i == twopowk)
		{
			k--;
			twopowk>>=1;
		}
	}

	// we can not encode anything with the very last element 
	permutation[nelements-1] = reference[0];

	// return number of bits encoded
	return bit_count - bit_count_start;
}

static int decode(Permute** reference, Permute** permutation, int nelements)
{
	int bit_count_start = bit_count;

	// compute k and when k will decrement
	int i = nelements;
	int k = 0;
	while (i/=2) k++;
	int twopowk = 1<<k;

	// decode either k+1 or k bits with each face
	int p, p_orig;
  for (i = 0; i < nelements-1; i++)
	{
		p = permutation[i]->p;
		if (p >= twopowk)
		{
			// we were able to decode k+1 bits
			p_orig = peek_bits(k+1);
			pop_bits(k+1);
		}
		else
		{
			// can only decode k bits
			p_orig = peek_bits(k);
			pop_bits(k);
		}
		// check that we decoded the right number
		if (p != p_orig)
		{
			fprintf(stderr,"decoded %d instead of %d\n",p,p_orig);
		}
    // these two lines replace the binary tree
		// the last face of the reference array becomes the new reference face p 
		reference[p] = reference[nelements-1-i];
		reference[p]->p = p;
		// is k becomming smaller
		if (nelements-i == twopowk)
		{
			k--;
			twopowk>>=1;
		}
	}
	// we can not decode anything from the very last face 
	if (permutation[nelements-1]->p != 0)
	{
		fprintf(stderr,"last face has position %d instead of 0\n",permutation[nelements-1]->p);
	}

	// return number of bits decoded
	return bit_count - bit_count_start;
}

int main(int argc, char **argv)
{
  int i;
  clock_t clock_start, clock_finish;

  if (argc != 2)
  {
    fprintf(stderr,"usage:\n");
    fprintf(stderr,"%s mesh.ply\n", argv[0]);
    exit(1);
  }

  // load the mesh

  int nverts;
  Vertex* verts;
  int nfaces;
  Face* faces;

  // start clock
  clock_start = clock();

  load_mesh(argv[1], &verts, &nverts, &faces, &nfaces);

  // stop clock
  clock_finish = clock();
  fprintf(stderr,"      loading mesh took %.3f seconds\n", (double)(clock_finish-clock_start)/CLOCKS_PER_SEC);

  int* twinorigin;
  int* twininv;

  // start clock
  clock_start = clock();

  // construct connectivity

  construct_connectivity(nfaces, faces, nverts, &twinorigin,  &twininv);

  // stop clock
  clock_finish = clock();
  fprintf(stderr,"      constructing connectivity took %.3f seconds\n", (double)(clock_finish-clock_start)/CLOCKS_PER_SEC);

  Vertex** verts_reference = new Vertex*[nverts];
  Face** faces_reference = new Face*[nfaces];

  // start clock
  clock_start = clock();

  // compute reference ordering
  compute_reference_ordering(nverts, verts, verts_reference, nfaces, faces, faces_reference, twinorigin, twininv);

  // stop clock
  clock_finish = clock();
  fprintf(stderr,"      computing reference ordering took %.3f seconds\n", (double)(clock_finish-clock_start)/CLOCKS_PER_SEC);

	// create arrays to store the computed permutation
  Vertex** verts_permutation = new Vertex*[nverts];
  Face** faces_permutation = new Face*[nfaces];

	// prepare the bits that are encoded
	init_bits();

  // start clock
  clock_start = clock();

	// encode as many bits as possible by permuting the vertices
	i = encode((Permute**)verts_reference, (Permute**)verts_permutation, nverts);
	fprintf(stderr,"encoded %d bits in %d vertices (avg %.2f bits)\n",i,nverts,((float)i)/((float)nverts));

	// encode as many bits as possible by permuting the faces
	i = encode((Permute**)faces_reference, (Permute**)faces_permutation, nfaces);
	fprintf(stderr,"encoded %d bits in %d faces (avg %.2f bits)\n",i,nfaces,((float)i)/((float)nfaces));

  // stop clock
  clock_finish = clock();

	// done encoding
	fprintf(stderr,"done. message of %d bits hidden in connectivity (%.2f b/v)\n",bit_count,((float)bit_count)/((float)nverts));
  fprintf(stderr,"      hiding message in mesh took %.3f seconds\n", (double)(clock_finish-clock_start)/CLOCKS_PER_SEC);

	// start decoding

	// re-create the same reference orderings that were created above
  delete twinorigin;
  delete twininv;
  construct_connectivity(nfaces, faces, nverts, &twinorigin,  &twininv);
  compute_reference_ordering(nverts, verts, verts_reference, nfaces, faces, faces_reference, twinorigin, twininv);

	// re-prepare the bits to check that we decode correctly
	init_bits();

  // start clock
  clock_start = clock();

  for (i = 0; i < nverts; i++)
  {
    verts_reference[i]->p = i;
  }

  for (i = 0; i < nfaces; i++)
  {
    faces_reference[i]->p = i;
  }

  // decode bits from vertex permutation
	i = decode((Permute**)verts_reference, (Permute**)verts_permutation, nverts);
	fprintf(stderr,"decoded %d bits from %d vertices\n",i,nverts);


  // decode bits from face permutation
	i = decode((Permute**)faces_reference, (Permute**)faces_permutation, nfaces);
	fprintf(stderr,"decoded %d bits from %d faces\n",i,nfaces);

  // stop clock
  clock_finish = clock();

	// done decoding
	fprintf(stderr,"done. message of %d bits recovered from connectivity\n",bit_count);
  fprintf(stderr,"      recovering and checking message took %.3f seconds\n", (double)(clock_finish-clock_start)/CLOCKS_PER_SEC);

  return 1;
}
